''' Run a quick example pipeline and see if the
    pipeline can be loaded and saved correctly
    with Marmalade
'''

import os
import shutil
import unittest
import tempfile
import autograd.numpy as np

from rlstm.pipe_save import Marmalade
from rlstm.pipeline import RLSTMPipeline
from rlstm.common_util import is_numpy_array, is_function
import rlstm.datasets

class TestMarmalade(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestMarmalade, cls).setUpClass()

        pipeline = RLSTMPipeline(
            os.path.abspath(os.path.dirname(rlstm.datasets.__file__)))
        pipeline.add_node('gru-hmm',
                          'gru-hmm',
                          {'hmm_state_count':5,
                           'gru_state_count':5,
                           'min_iters':1,
                           'max_iters':5,
                           'batch_size':64,
                           'param_scale':0.01,
                           'stop_criterion':1e-3,
                           'early_stop':False,
                           'has_va_set':False})

        pipeline.run_and_eval()
        cls.pipe = pipeline
        cls.node = pipeline.get_node(index=0)

        if not os.path.isdir('_marmalade_tests_folder'):
            os.mkdir('_marmalade_tests_folder')

    @classmethod
    def tearDownClass(cls):
        if os.path.isdir('_marmalade_tests_folder'):
            shutil.rmtree('_marmalade_tests_folder')

    def test_marmalade_pipeline(self):
        node = self.node
        pipe = self.pipe

        with open('pipe.marma', 'w') as fp:
            Marmalade.serialize_pipe(pipe, fp)

        with open('pipe.marma') as fp:
            reload_pipe = Marmalade.unserialize_pipe(fp)
            reload_node = reload_pipe.get_node(index=0)

            self.assertEqual(reload_pipe.num_nodes, pipe.num_nodes)
            self.assertEqual(reload_pipe.log_file, pipe.log_file)
            self.assertEqual(reload_pipe.weights_file, pipe.weights_file)
            self.assertEqual(reload_pipe.dataset, pipe.dataset)
            self.assertEqual(reload_pipe.obs_count, pipe.obs_count)
            self.assertEqual(reload_pipe.out_count, pipe.out_count)

            self.assertEqual(reload_node.name, node.name)
            self.assertEqual(reload_node.index, node.index)
            self.assertEqual(reload_node.prev, node.prev)
            self.assertEqual(reload_node.next, node.next)
            self.assertEqual(reload_node.type, node.type)

            for param in reload_node.params:
                if is_numpy_array(reload_node.params[param]):
                    self.assertEqual(
                        np.sum(~(reload_node.params[param] == node.params[param])), 0)
                elif not is_function(reload_node.params[param]):
                    self.assertEqual(reload_node.params[param], node.params[param])
                else:  # function
                    self.assertEqual(reload_node.params[param].__code__.co_code,
                                     node.params[param].__code__.co_code)

    def test_marmalade_node(self):
        node = self.node

        with open('_marmalade_tests_folder/node.marma', 'w') as fp:
            Marmalade.serialize_node(self.node, fp)

        with open('_marmalade_tests_folder/node.marma') as fp:
            reload_node = Marmalade.unserialize_node(fp)

            self.assertEqual(reload_node.name, node.name)
            self.assertEqual(reload_node.index, node.index)
            self.assertEqual(reload_node.prev, node.prev)
            self.assertEqual(reload_node.next, node.next)
            self.assertEqual(reload_node.type, node.type)

            for param in reload_node.params:
                if is_numpy_array(reload_node.params[param]):
                    self.assertEqual(
                        np.sum(~(reload_node.params[param] == node.params[param])), 0)
                elif not is_function(reload_node.params[param]):
                    self.assertEqual(reload_node.params[param], node.params[param])
                else:  # function
                    self.assertEqual(reload_node.params[param].__code__.co_code,
                                     node.params[param].__code__.co_code)



from __future__ import absolute_import, print_function

from builtins import range
from copy import copy

import autograd.numpy as np
from autograd import value_and_grad
from autograd.scipy.misc import logsumexp

from rlstm.models import gru, grudrop, gruredun, hmm
from rlstm.weights_parser import WeightsParser
from rlstm.nn_util import sigmoid, add_bias_and_multiply
from rlstm.common_util import safe_log


def normalize(x):
    norm = np.linalg.norm(x)
    if norm == 0:
        return x
    return x / norm


def kdropout(weights, k):
    ''' chooses k indexes of x. set others to 0. '''
    num_inputs = weights.shape[0]
    num_states = weights.shape[1]
    new_weights = copy(weights)
    # sort the weights
    sort_id_weights = np.argsort(np.abs(new_weights), axis=1)
    sort_weights = new_weights[np.arange(num_inputs)[:, None], sort_id_weights]
    sort_weights = np.hstack(
        (np.zeros((num_inputs, num_states-k)), sort_weights[:, -k:]))
    # recover indexes into new_weights
    recover_sort_id = np.argsort(sort_id_weights, axis=1)
    new_weights = sort_weights[np.arange(num_inputs)[:, None], recover_sort_id]
    return new_weights


def row_norms(X):
    norms = np.einsum('ij,ij->i', X, X)
    return np.sqrt(norms)


def cosine_cdist(XA, XB):
    mA = XA.shape[0]
    mB = XB.shape[0]
    dm = np.zeros((mA, mB), dtype=np.double)

    normsA = row_norms(XA)
    normsB = row_norms(XB)

    dm = np.dot(XA, XB.T)
    dm /= normsA.reshape(-1, 1)
    dm /= normsB
    dm *= -1
    dm += 1
    return dm


def sparse_cos_dists(A, B):
    cos_dists_output = cosine_cdist(A, B)
    # symmetric, so just keep 1/2 (ignore diagonal)
    return np.triu(cos_dists_output, k=1)


def is_redundant(A, epsilon=0.5):
    redun_idx = []
    idx = np.argwhere(np.logical_and(np.abs(A) < epsilon, A != 0))
    if idx.shape[0] > 0:  # get unique so no duplicate removals
        # quick way to get rid of duplicates
        idx = idx[np.unique(idx[:, 1], return_index=True)[1], :]
        # create a map from uid1 to uid2
        uid_pool = np.unique(idx[:, 0])
        for uid in uid_pool:
            redun_idx.append((uid, idx[:, 1][idx[:, 0] == uid]))
    return redun_idx, idx[:, 1]


def autograd_zero(A, index):
    ''' Because autograd can't support indexing, we have to
        be kind of clever about it
            - this handles empty indexes
    '''
    B = copy(A.T)
    if index.shape[0] > 0:
        index_all = np.arange(B.shape[1])
        index_ordered = np.concatenate(
            (index, index_all[~np.in1d(index_all, index)]))
        B_ordered = B[np.arange(B.shape[0])[:, None], index_ordered]
        B_zeroed = np.concatenate(
            (np.zeros((B.shape[0], len(index))), B_ordered[:, len(index):]), axis=1)
        index_recover = np.argsort(index_ordered)
        B = B_zeroed[np.arange(B.shape[0])[:, None], index_recover]
    return B.T


def autograd_add(A, index):
    B = copy(A)
    if len(index) > 0:
        for i, j in index:
            C = np.expand_dims(B[i, :] + np.sum(B[j, :], axis=0), axis=0)
            B = np.concatenate((B[:i, :], C, B[i+1:, :]))
    return B


def split_weights(weights, num_hiddens):
    num_parts = weights.shape[1] / num_hiddens
    return [weights[:, i*num_hiddens:(i+1)*num_hiddens]
            for i in range(num_parts)]


def hiddens_to_output_probs(
        hiddens,
        output_h_weights,
        lstate_norm=0,
        output_hmm_weights=0):

    state = np.vstack((copy(lstate_norm), np.ones((1, lstate_norm.shape[1]))))
    output = sigmoid(np.dot(hiddens, output_h_weights) +
                     np.transpose(np.dot(output_hmm_weights, state)))
    # output = logsumexp( output ) # normalize log-probs (multinomial case)
    return safe_log(output), safe_log(1 - output)


def update_gru(
        max_conn,
        redun_epsilon,
        curr_input,
        prev_hiddens,
        update_x_weights,
        update_h_weights,
        reset_x_weights,
        reset_h_weights,
        thidden_x_weights,
        thidden_h_weights):

    if redun_epsilon > 0:
        num_hiddens = prev_hiddens.shape[1]

        # get "redundant" index by cosine distances between weight matrices
        all_x_weights = np.concatenate(
            (update_x_weights, reset_x_weights, thidden_x_weights), axis=1)
        all_x_redun, xy_redun = is_redundant(
            sparse_cos_dists(all_x_weights, all_x_weights), redun_epsilon)

        all_h_weights = np.concatenate(
            (update_h_weights, reset_h_weights, thidden_h_weights), axis=1)
        all_h_redun, hy_redun = is_redundant(
            sparse_cos_dists(all_h_weights, all_h_weights), redun_epsilon)

        # 0 out the proper indexes in the connections
        all_x_weights = autograd_zero(
            autograd_add(all_x_weights, all_x_redun), xy_redun)
        all_h_weights = autograd_zero(
            autograd_add(all_h_weights, all_h_redun), hy_redun)

        # get the individual vectors out
        update_x_weights, reset_x_weights, thidden_x_weights = \
            split_weights(all_x_weights, num_hiddens)
        update_h_weights, reset_h_weights, thidden_h_weights = \
            split_weights(all_h_weights, num_hiddens)

    if max_conn > 0:
        update_x_weights = kdropout(update_x_weights, max_conn)
        update_h_weights = kdropout(update_h_weights, max_conn)
        reset_x_weights = kdropout(reset_x_weights, max_conn)
        reset_h_weights = kdropout(reset_h_weights, max_conn)
        thidden_x_weights = kdropout(thidden_x_weights, max_conn)
        thidden_h_weights = kdropout(thidden_h_weights, max_conn)

    update = sigmoid(
        add_bias_and_multiply(update_x_weights, curr_input) +
        np.dot(prev_hiddens, update_h_weights))

    reset = sigmoid(
        add_bias_and_multiply(reset_x_weights, curr_input) +
        np.dot(prev_hiddens, reset_h_weights))

    thiddens = np.tanh(
        add_bias_and_multiply(thidden_x_weights, curr_input) +
        np.dot(reset * prev_hiddens, thidden_h_weights))

    hiddens = (1 - update) * prev_hiddens + update * thiddens
    return hiddens

# Update GRU and HMM parameters independently


def update(
        max_conn,
        redun_epsilon,
        curr_input,
        prev_hiddens,
        update_x_weights,
        update_h_weights,
        reset_x_weights,
        reset_h_weights,
        thidden_x_weights,
        thidden_h_weights,
        prev_lstate,
        ltrans_mat,
        lobs_mat):

    hiddens = update_gru(max_conn, redun_epsilon,
                         curr_input, prev_hiddens, update_x_weights,
                         update_h_weights, reset_x_weights, reset_h_weights,
                         thidden_x_weights, thidden_h_weights)
    curr_lstate = hmm.update(curr_input, prev_lstate, ltrans_mat, lobs_mat)
    return hiddens, curr_lstate


def build(
        input_count,
        hmm_state_count,
        gru_state_count,
        output_count,
        max_conn=0,
        redun_epsilon=0):

    # the input count is just the number of input dimensions, we
    # assume that any/all 1-hot encoding, etc. has been done already.
    # the output count is the *number* of *binary* outputs (no
    # multinomial or continuous support at the moment)
    parser = WeightsParser()

    # GRU parameters
    parser.add_shape('init_hiddens', (1, gru_state_count))
    parser.add_shape('update_x_weights', (input_count + 1, gru_state_count))
    parser.add_shape('update_h_weights', (gru_state_count, gru_state_count))
    parser.add_shape('reset_x_weights', (input_count + 1, gru_state_count))
    parser.add_shape('reset_h_weights', (gru_state_count, gru_state_count))
    parser.add_shape('thidden_x_weights', (input_count + 1, gru_state_count))
    parser.add_shape('thidden_h_weights', (gru_state_count, gru_state_count))
    parser.add_shape('output_h_weights', (gru_state_count, output_count))

    num_gru_weights = parser.num_weights

    # HMM parameters
    parser.add_shape('lpi_mat', (hmm_state_count, 1))
    parser.add_shape('ltrans_mat', (hmm_state_count, hmm_state_count))
    parser.add_shape('lobs_mat', (hmm_state_count, input_count))
    parser.add_shape('output_hmm_weights', (output_count, hmm_state_count + 1))

    num_hmm_weights = parser.num_weights - num_gru_weights

    # Make predictions through a linear combination of GRU and HMM parameters
    def outputs(weights, input_set):
        update_x_weights = parser.get(weights, 'update_x_weights')
        update_h_weights = parser.get(weights, 'update_h_weights')
        reset_x_weights = parser.get(weights, 'reset_x_weights')
        reset_h_weights = parser.get(weights, 'reset_h_weights')
        thidden_x_weights = parser.get(weights, 'thidden_x_weights')
        thidden_h_weights = parser.get(weights, 'thidden_h_weights')
        output_h_weights = parser.get(weights, 'output_h_weights')

        lpi_mat = parser.get(weights, 'lpi_mat')
        ltrans_mat = parser.get(weights, 'ltrans_mat')
        lobs_mat = parser.get(weights, 'lobs_mat')

        output_hmm_weights = parser.get(weights, 'output_hmm_weights')

        lpi_mat = lpi_mat - logsumexp(lpi_mat, axis=0)
        ltrans_mat = ltrans_mat - logsumexp(ltrans_mat, axis=1, keepdims=True)

        data_count = input_set.shape[2]
        time_count = input_set.shape[1]

        # the output set of dimension ( output_count, time_count, data_count )
        # output_set = np.zeros( ( output_count, time_count, data_count ) )
        all_ll = 0
        for data_iter in range(data_count):

            # Initialize hidden states in the HMM
            hiddens = copy(parser.get(weights, 'init_hiddens'))

            # Initialize belief in HMM with the first set of observations so
            # things are sync'd appropriately
            lstate = copy(lpi_mat)
            my_input = copy(input_set[:, 0, data_iter])
            my_input[my_input < .5] = -1
            tmp = -1 * \
                logsumexp(
                    hmm.my_dstack([np.zeros(lobs_mat.shape), lobs_mat * my_input]), axis=2)
            lstate = lstate + np.sum(tmp, 1, keepdims=True)

            # Loop over time steps
            for time_iter in range(time_count):
                hiddens, lstate = update(
                    max_conn,
                    redun_epsilon,
                    np.expand_dims(input_set[:, time_iter, data_iter], axis=0),
                    hiddens,
                    update_x_weights,
                    update_h_weights,
                    reset_x_weights,
                    reset_h_weights,
                    thidden_x_weights,
                    thidden_h_weights,
                    lstate,
                    ltrans_mat,
                    lobs_mat)

                lstate_norm = lstate - logsumexp(lstate)
                # output_set[:,time_iter,data_iter] = \
                #    hiddens_to_output_probs( hiddens, output_h_weights )
                # more convoluted way with hstack and vstack because
                # autograd does not support assigning with indices
                out_vec1, out_vec0 = hiddens_to_output_probs(
                    hiddens,
                    output_h_weights,
                    lstate_norm,
                    output_hmm_weights)

                out_vec1 = np.transpose(out_vec1)
                out_vec0 = np.transpose(out_vec0)
                if time_iter == 0:
                    out_mat1 = out_vec1
                    out_mat0 = out_vec0
                    lmat = lstate_norm
                else:
                    out_mat1 = np.hstack((out_mat1, out_vec1))
                    out_mat0 = np.hstack((out_mat0, out_vec0))
                    lmat = np.hstack((lmat, lstate_norm))

            my_ll = logsumexp(lstate)
            all_ll = all_ll + my_ll
            if data_iter == 0:
                output_set1 = out_mat1
                output_set0 = out_mat0
                lstate_set = lmat
            else:
                output_set1 = np.dstack((output_set1, out_mat1))
                output_set0 = np.dstack((output_set0, out_mat0))
                lstate_set = np.dstack((lstate_set, lmat))

        return output_set1, output_set0, all_ll, lstate_set

    def prediction(weights, input_set):
        out, _, _, _ = outputs(weights, input_set)
        return out

    def log_likelihood(weights, input_set, output_set=None):
        lprobs1, lprobs0, all_ll, lstate_set = outputs(weights, input_set)
        if output_set is None:
            return all_ll
        else:
            lprobs = output_set * lprobs1 + (1 - output_set) * lprobs0
            ll = np.sum(lprobs)
            return ll

    return prediction, log_likelihood, num_gru_weights, num_hmm_weights



from __future__ import absolute_import, print_function

import pdb
from builtins import range
from copy import copy

import autograd.numpy as np
from autograd import value_and_grad
from autograd.scipy.misc import logsumexp

from rlstm.weights_parser import WeightsParser
from rlstm.nn_util import sigmoid, add_bias_and_multiply
from rlstm.common_util import safe_log


def row_norms(X):
    norms = np.einsum('ij,ij->i', X, X)
    return np.sqrt(norms)


def cosine_cdist(XA, XB):
    mA = XA.shape[0]
    mB = XB.shape[0]
    dm = np.zeros((mA, mB), dtype=np.double)

    normsA = row_norms(XA)
    normsB = row_norms(XB)

    dm = np.dot(XA, XB.T)
    dm /= normsA.reshape(-1, 1)
    dm /= normsB
    dm *= -1
    dm += 1
    return dm


def sparse_cos_dists(A, B):
    cos_dists_output = cosine_cdist(A, B)
    # symmetric, so just keep 1/2 (ignore diagonal)
    return np.triu(cos_dists_output, k=1)


def is_redundant(A, epsilon=0.5):
    redun_idx = []
    idx = np.argwhere(np.logical_and(np.abs(A) < epsilon, A != 0))
    if idx.shape[0] > 0:  # get unique so no duplicate removals
        # quick way to get rid of duplicates
        idx = idx[np.unique(idx[:, 1], return_index=True)[1], :]
        # create a map from uid1 to uid2
        uid_pool = np.unique(idx[:, 0])
        for uid in uid_pool:
            redun_idx.append((uid, idx[:, 1][idx[:, 0] == uid]))
    return redun_idx, idx[:, 1]


def autograd_zero(A, index):
    ''' Because autograd can't support indexing, we have to
        be kind of clever about it
            - this handles empty indexes
    '''
    B = copy(A.T)
    if index.shape[0] > 0:
        index_all = np.arange(B.shape[1])
        index_ordered = np.concatenate(
            (index, index_all[~np.in1d(index_all, index)]))
        B_ordered = B[np.arange(B.shape[0])[:, None], index_ordered]
        B_zeroed = np.concatenate(
            (np.zeros((B.shape[0], len(index))), B_ordered[:, len(index):]), axis=1)
        index_recover = np.argsort(index_ordered)
        B = B_zeroed[np.arange(B.shape[0])[:, None], index_recover]
    return B.T


def autograd_add(A, index):
    B = copy(A)
    if len(index) > 0:
        for i, j in index:
            C = np.expand_dims(B[i, :] + np.sum(B[j, :], axis=0), axis=0)
            B = np.concatenate((B[:i, :], C, B[i+1:, :]))
    return B


def split_weights(weights, num_hiddens):
    num_parts = weights.shape[1] / num_hiddens
    return [weights[:, i*num_hiddens:(i+1)*num_hiddens]
        for i in range(num_parts)]


def update(
        curr_input,
        prev_hiddens,
        update_x_weights,
        update_h_weights,
        reset_x_weights,
        reset_h_weights,
        thidden_x_weights,
        thidden_h_weights,
        redun_epsilon=0.5):

    ''' breaks connections not in the nodes but in the wiring '''
    num_hiddens = prev_hiddens.shape[1]

    # get "redundant" index by cosine distances between weight matrices
    all_x_weights = np.concatenate(
        (update_x_weights, reset_x_weights, thidden_x_weights), axis=1)
    all_x_redun, xy_redun = is_redundant(
        sparse_cos_dists(all_x_weights, all_x_weights), redun_epsilon)

    all_h_weights = np.concatenate(
        (update_h_weights, reset_h_weights, thidden_h_weights), axis=1)
    all_h_redun, hy_redun = is_redundant(
        sparse_cos_dists(all_h_weights, all_h_weights), redun_epsilon)

    # 0 out the proper indexes in the connections
    all_x_weights = autograd_zero(
        autograd_add(all_x_weights, all_x_redun), xy_redun)
    all_h_weights = autograd_zero(
        autograd_add(all_h_weights, all_h_redun), hy_redun)

    # get the individual vectors out
    update_x_weights, reset_x_weights, thidden_x_weights = \
        split_weights(all_x_weights, num_hiddens)
    update_h_weights, reset_h_weights, thidden_h_weights = \
        split_weights(all_h_weights, num_hiddens)

    # proceed as regular gru
    update = sigmoid(
        add_bias_and_multiply(update_x_weights, curr_input) +
        np.dot(prev_hiddens, update_h_weights))

    reset = sigmoid(
        add_bias_and_multiply(reset_x_weights, curr_input) +
        np.dot(prev_hiddens, reset_h_weights))

    thiddens = np.tanh(
        add_bias_and_multiply(thidden_x_weights, curr_input) +
        np.dot(reset * prev_hiddens, thidden_h_weights))

    hiddens = (1 - update) * prev_hiddens + update * thiddens

    return hiddens


def hiddens_to_output_probs(hiddens, output_h_weights):
    # binary case, each a prob
    output = sigmoid(np.dot(hiddens, output_h_weights))
    # output - logsumexp( output ) # normalize log-probs (multinomial case)
    return safe_log(output), safe_log(1 - output)


def build(input_count, state_count, output_count, redun_epsilon=0.5):
    # the input count is just the number of input dimensions, we
    # assume that any/all 1-hot encoding, etc. has been done already.
    # the output count is the *number* of *binary* outputs (no
    # multinomial or continuous support at the moment)

    parser = WeightsParser()
    parser.add_shape('init_hiddens', (1, state_count))
    parser.add_shape('update_x_weights', (input_count + 1, state_count))
    parser.add_shape('update_h_weights', (state_count, state_count))
    parser.add_shape('reset_x_weights', (input_count + 1, state_count))
    parser.add_shape('reset_h_weights', (state_count, state_count))
    parser.add_shape('thidden_x_weights', (input_count + 1, state_count))
    parser.add_shape('thidden_h_weights', (state_count, state_count))
    parser.add_shape('output_h_weights', (state_count, output_count))

    def outputs(weights, input_set):
        update_x_weights = parser.get(weights, 'update_x_weights')
        update_h_weights = parser.get(weights, 'update_h_weights')
        reset_x_weights = parser.get(weights, 'reset_x_weights')
        reset_h_weights = parser.get(weights, 'reset_h_weights')
        thidden_x_weights = parser.get(weights, 'thidden_x_weights')
        thidden_h_weights = parser.get(weights, 'thidden_h_weights')
        output_h_weights = parser.get(weights, 'output_h_weights')

        data_count = input_set.shape[2]
        time_count = input_set.shape[1]

        # the output set of dimension ( output_count, time_count, data_count )
        # output_set = np.zeros( ( output_count, time_count, data_count ) )
        for data_iter in range(data_count):
            hiddens = copy(parser.get(weights, 'init_hiddens'))
            for time_iter in range(time_count):
                hiddens = update(np.expand_dims(input_set[:, time_iter, data_iter], axis=0),
                                          hiddens, update_x_weights, update_h_weights,
                                          reset_x_weights, reset_h_weights,
                                          thidden_x_weights, thidden_h_weights,
                                          redun_epsilon=redun_epsilon)

                # output_set[:,time_iter,data_iter] = \
                #    hiddens_to_output_probs( hiddens, output_h_weights )
                # more convoluted way with hstack and vstack because
                # autograd does not support assigning with indices
                out_vec1, out_vec0 = hiddens_to_output_probs(
                    hiddens, output_h_weights)
                out_vec1 = np.transpose(out_vec1)
                out_vec0 = np.transpose(out_vec0)
                if time_iter == 0:
                    out_mat1 = out_vec1
                    out_mat0 = out_vec0
                else:
                    out_mat1 = np.hstack((out_mat1, out_vec1))
                    out_mat0 = np.hstack((out_mat0, out_vec0))

            if data_iter == 0:
                output_set1 = out_mat1
                output_set0 = out_mat0
            else:
                output_set1 = np.dstack((output_set1, out_mat1))
                output_set0 = np.dstack((output_set0, out_mat0))
        return output_set1, output_set0

    def prediction(weights, input_set):
        out, _ = outputs(weights, input_set)
        return out

    def log_likelihood(weights, input_set, output_set):
        lprobs1, lprobs0 = outputs(weights, input_set)
        lprobs = output_set * lprobs1 + (1 - output_set) * lprobs0
        ll = np.sum(lprobs)
        return ll

    return prediction, log_likelihood, parser.num_weights



import os
import sys

import numpy as np

# Convert from the 2D data format (subjects x time points, observations) described here
# https://gist.github.com/michaelchughes/76336975162477dfb4b38ffd881da3d5
# to 3D data format (observations, time points, subjects)

def convert_single_patient_data(dir_name = 'physionet2014/1009'):
  import h5py
  times = {'train': 150000, 'test': 66000}
  suff = {'train' : '', 'test': '_val'}
  for data_type in ['train', 'test']:
    f = h5py.File(dir_name+'/data'+suff[data_type]+'.h5','r')
    x = np.transpose(f['obs'])
    x = x.reshape(6,times[data_type],1)
    np.save(dir_name+'_3D/X_'+data_type+'.npy',x)
    y = np.reshape(f['output'], (1,times[data_type],1))
    np.save(dir_name+'_3D/y_'+data_type+'.npy',y)
    f.close()

directory = sys.argv[1]
if not os.path.exists(directory+'_3D'):
  os.mkdir(directory+'_3D')

for data_type in ['train', 'test']:
    X = np.load(directory+'/X_'+data_type+'.npy')
    Y = np.load(directory+'/Y_'+data_type+'.npy')

    fence = np.load(directory+'/fenceposts_'+data_type+'.npy')
    max_time = max(fence[1:] - fence[:-1])
    num_subjects = len(fence)-1

    obs_set = np.zeros((X.shape[1], max_time, num_subjects))
    out_set = np.zeros((Y.shape[1], max_time, num_subjects))

    for i in range(num_subjects):
        time = fence[i+1] - fence[i]
        obs_set[:,:time,i] = X[fence[i]:fence[i+1]].transpose()
        out_set[:,:time,i] = Y[fence[i]:fence[i+1]].transpose()

    np.save(directory+'_3D/X_'+data_type, obs_set)
    np.save(directory+'_3D/y_'+data_type, out_set)



"""
Similar file to early_stopping/optimize.py but
depends on converging vs early_stopping
"""

from __future__ import absolute_import

import autograd.numpy as np
from autograd.util import flatten_func
from builtins import range
from random import sample
from math import ceil
from copy import copy
from rlstm.scores import get_accuracy


def adam(
        grad,
        init_params,
        callback=None,
        min_iters=0,
        max_iters=1e5,
        step_size=0.001,
        b1=0.9,
        b2=0.999,
        eps=10**-8,
        ll_fun=None,
        va_ll_fun=None,
        stop_criterion=1e-3,
        early_stop=False,
        patience=10):
    """ Adam as described in http://arxiv.org/pdf/1412.6980.pdf.
    It's basically RMSprop with momentum and some correction terms."""

    flattened_grad, unflatten, x = flatten_func(grad, init_params)

    # initial settings for variables
    m, v = np.zeros(len(x)), np.zeros(len(x))
    cur_iter = 0
    reset_patience = patience
    old_ll, cur_ll, cur_va_ll, best_va_ll = 1, 0, None, None

    while ((cur_iter < max_iters) and (np.abs(cur_ll - old_ll) > stop_criterion)) \
            or (cur_iter < min_iters):

        g = flattened_grad(x, cur_iter)  # pass iter for batch training
        m = (1 - b1) * g + b1 * m  # First  moment estimate.
        v = (1 - b2) * (g**2) + b2 * v  # Second moment estimate.
        mhat = m / (1 - b1**(cur_iter + 1))    # Bias correction.
        vhat = v / (1 - b2**(cur_iter + 1))
        x = x - step_size*mhat/(np.sqrt(vhat) + eps)

        if ll_fun:
            old_ll = cur_ll
            cur_ll = ll_fun(x)

        if va_ll_fun:
            cur_va_ll = va_ll_fun(x)

            # stop based on va ll
            if early_stop:
                if cur_iter == 0:
                    best_va_ll = cur_va_ll
                    best_x = x
                else:
                    if cur_va_ll <= best_va_ll:
                        best_va_ll = cur_va_ll
                        best_x = x
                        patience = reset_patience
                    else:
                        patience -= 1

                if patience <= 0:
                    break

        if callback:
            callback(cur_iter, cur_ll, cur_va_ll, patience if early_stop else None, x)

        cur_iter += 1
    return unflatten(best_x if early_stop else x)




from __future__ import print_function

import autograd.numpy as np
from sklearn.metrics import auc, roc_curve

from rlstm.common_util import flatten_to_2d, flatten_to_2d


def get_auc(outputs, probas):
    ''' AUC is a common metric for binary classification
        methods by comparing true & false positive rates

        Args
        ----
        outputs : numpy array of true outcomes (OxTxN)
        probas  : numpy array of predicted probabilities (OxTxN)
    '''
    fpr, tpr, _ = roc_curve(outputs, probas)
    return auc(fpr, tpr)


def get_accuracy(outputs, probas):
    return np.mean(outputs.flatten() == np.round(probas, 0).flatten())


def get_mean_squared_error(outputs, probas):
    return np.mean((outputs.flatten() - probas.flatten())**2)


def get_log_sparsity(weights, thres=1e-3):
    ''' Sparser weights are more interpretable.
        This returns a fraction of weights that are below
        a threshold.

        Args
        ----
        weights : numpy array of weights (1xW)
        thres   : float
    '''
    nonzero_count = np.sum(np.abs(weights) < thres)
    all_count = float(weights.shape[0])
    return nonzero_count / all_count


def get_pr_obs(ll_fun, weights, obs):
    ''' Given that it is important to model observations,
        not just outputs, then this tells us how well we
        can explain the observations.

        Args
        ----
        ll_fun  : function provided by one of the models in
                          this repo. Returns a lambda function
        weights : numpy array of weights (1xW)
        obs     : numpy array of inputs (IxTxN)
    '''
    return ll_fun(weights, obs)



import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import grad, value_and_grad

from rlstm.common_util import safe_log
from rlstm.nn_util import sigmoid, flatten_io


def logistic_predictions(weights, input_set, in_dim_count, out_dim_count):
    # input should be ( dims by data ) or vertical columns
    # Adds in the row of ones, so we don't have to when passing things in!
    # outputs probability of a label being true according to logistic model.
    if len(input_set.shape) > 2:
        input_set = flatten_io(input_set)
    my_input_set = np.vstack((input_set, np.ones((1, input_set.shape[1]))))
    weights = np.reshape(weights, (out_dim_count, in_dim_count + 1))
    return sigmoid(np.dot(weights, my_input_set))


def log_likelihood(weights, input_set, output_set, debug=False):
    # Training loss is the negative log-likelihood of the training labels.
    if len(input_set.shape) > 2:
        input_set, output_set = flatten_io(input_set, output_set)
    in_dim_count = input_set.shape[0]
    out_dim_count = output_set.shape[0]
    preds = logistic_predictions(
        weights, input_set, in_dim_count, out_dim_count)
    label_probabilities = safe_log(
        preds) * output_set + safe_log(1 - preds) * (1 - output_set)
    if debug:
        return np.sum(label_probabilities), label_probabilities
    return np.sum(label_probabilities)



import os
from setuptools import setup, find_packages

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

setup(name='rlstm',
      packages=['rlstm'],
      package_dir={'rlstm': 'rlstm'},
      package_data={'rlstm': ['models/*.py',
                              'compression/*.py',
                              'distillation/*.py',
                              'datasets/*',
                              'tests/*.py',
                              'interpret/*.py']},
      version='0.1.37',
      description=u'Pipeline tool for machine learning models.',
      keywords=['generative', 'interpretable', 'hmm', 'lstm', 'residual'],
      author=u'Mike Wu',
      author_email='me@mikewuis.me',
      url='https://github.com/dtak/interpretable-models',
      download_url='https://github.com/dtak/interpretable-models/tarball/0.1.0',
      classifiers=[],
      install_requires=[
          'argh==0.26.2',
          'numpy==1.11.2',
          'scipy==0.18.1',
          'scikit-learn==0.18',
          'autograd==1.1.6',
          'pydot==1.2.3',
          'pydotplus==2.0.2',
      ],
      zip_safe=False
    )
